import re
import warnings

from os import path
from functools import cache

from .content import grammars_dir, total_grammars_dir

from . import translation as hedy_translation

"""
Because of the gradual nature of Hedy, the grammar of every level is just slightly different than the grammar of the
previous one. With the current approach every level only describes the grammar differences from its preceding level.
To get the grammar of level N, the grammar of level 1 is merged consecutively with the change definitions of all
levels up to N. To facilitate this approach, 2 features are added:

    - Preprocessing rules appear in grammar rule definitions and allow for custom python logic to be applied before the
    rule is merged. For example, `for:<needs_colon>` fetches the `for` rule from the base grammar and adds a colon at
    the end of its definition. Another example is `if_error:<old_rule_to_error ifs>` which fetches the definition of
    `ifs` from the base grammar. Preprocessors are also used to construct regular expressions which have to avoid
    specific translated keywords, e.g. `elses: /([^\n ]| (?!<expand_keyword else>))+/` is transformed to
    `elses: /([^\n ]| (?!else|ellers))+/` in Danish and `elses: /([^\n ]| (?!else|değilse))+/` in Turkish.

    - Grammar mering operators, i.e. +=, -= and >>, allow adding, removing and moving last parts of a rule definition.
    For example, if level 1 contains the following definition `command: repeat | while` and level 2 redefines the
    rule as `command += for -= while >> repeat`, then the merged grammar of level 2 will have the following
    definition `command: for | repeat`.
"""


@cache
def create_grammar(level, lang, skip_faulty):
    """ Creates a grammar file for a chosen level and lang. Note that the language is required
    to generate regular expressions that escape keywords (with negative lookahead).
    Currently, it is only a couple of MB in total, so it is safe to cache. """
    # start with creating the grammar for level 1
    merged_grammars = get_full_grammar_for_level(1)

    # then keep merging new grammars in
    for lvl in range(2, level + 1):
        grammar_text_lvl = get_additional_rules_for_level(lvl)
        merged_grammars = merge_grammars(merged_grammars, grammar_text_lvl, lang)

    if skip_faulty:
        skip_faulty_grammar = read_skip_faulty_file(level)
        merged_grammars = merge_grammars(merged_grammars, skip_faulty_grammar, lang)

    # keyword and other terminals never have merge-able rules, so we can just add them at the end
    keywords = get_keywords_for_language(lang)
    terminals = get_terminals()
    merged_grammars = merged_grammars + '\n' + keywords + '\n' + terminals

    # ready? Save to file to ease debugging
    # this could also be done on each merge for performance reasons
    save_total_grammar_file(level, merged_grammars, lang)
    return merged_grammars


def merge_grammars(grammar_text_1, grammar_text_2, lang):
    """ Merges two grammar files into one.
    Rules that are redefined in the second file are overridden.
    Rules that are new in the second file are added."""
    merged_grammar = []
    rules_to_delete = []  # collects rules we no longer need

    base_grammar = parse_grammar(grammar_text_1)
    target_grammar = parse_grammar(grammar_text_2)

    apply_preprocessing_rules(target_grammar, base_grammar, lang)

    for base_rule in base_grammar.values():
        if base_rule.name in target_grammar:
            target_rule = target_grammar[base_rule.name]
            if base_rule.value.strip() == target_rule.value.strip():
                warnings.warn(f"The rule {base_rule.name} is duplicated: {base_rule.value}. Please check!")
            # computes the rules that use the merge operators in the grammar, namely +=, -= and >>
            merged_rule, to_delete = merge_rules_operator(base_rule.value, target_rule.value,
                                                          base_rule.name, target_rule.line)
            rules_to_delete.extend(to_delete)
            merged_grammar.append(merged_rule)
        else:
            merged_grammar.append(base_rule.line)

    for target_rule in target_grammar.values():
        if target_rule.name not in base_grammar:
            merged_grammar.append(target_rule.line)

    # filters rules that are no longer needed
    rules_to_keep = [rule for rule in merged_grammar if split_rule_name_and_value(rule)[0] not in rules_to_delete]
    return '\n'.join(sorted(rules_to_keep))


def read_grammars_file(*paths):
    path_ = path.join(grammars_dir(), *paths)
    with open(path_, "r", encoding="utf-8") as file:
        return file.read()


def read_skip_faulty_file(level):
    for lvl in range(level, 0, -1):
        file_path = path.join(grammars_dir(), f'skip-faulty-level{lvl}.lark')
        if path.isfile(file_path):
            with open(file_path, "r", encoding="utf-8") as file:
                return file.read()


def write_total_grammars_file(content, *paths):
    path_ = path.join(total_grammars_dir(), *paths)
    with open(path_, "w", encoding="utf-8") as file:
        file.write(content)


def get_keywords_for_language(language):
    try:
        return read_grammars_file(f'keywords-{language}.lark')
    except FileNotFoundError:
        return read_grammars_file('keywords-en.lark')


def get_terminals():
    return read_grammars_file('terminals.lark')


def save_total_grammar_file(level, grammar, lang_):
    write_total_grammars_file(grammar, f'level{level}.{lang_}-Total.lark')


def get_additional_rules_for_level(level):
    return read_grammars_file(f'level{level}-Additions.lark')


def get_full_grammar_for_level(level):
    return read_grammars_file(f'level{level}.lark')


def parse_grammar(grammar):
    lines = grammar.split('\n')
    rules = [parse_grammar_rule(line) for line in lines if line != '' and line[0] != '/']
    return {r.name: r for r in rules}


def parse_grammar_rule(line):
    processor_rules = [re.findall(fr'<({rule})( +[\w_]+)?>', line) for rule in PREPROCESS_RULES]
    processor_rules = [
        RuleProcessor(name, arg)
        for rules in processor_rules
        for name, arg in rules if rules
    ]

    name, value = split_rule_name_and_value(line)
    return GrammarRule(line=line, name=name, value=value, processors=processor_rules)


class GrammarRule:
    """Used to store information about parsed grammar rules when merging grammars.
    Valid examples of rules: name.-100: _DEFINITION, name<processor>, and name.1<processor argument>."""

    def __init__(self, line, name, value, processors=None):
        self.line = line
        self.name_with_priority = name
        self.name = strip_priority_suffix(name).strip()
        self.value = value
        self.processors = processors

    def apply_processors(self, base_grammar, lang):
        if self.processors:
            result = self.value
            for processor in self.processors:
                arg = processor.arg if processor.arg else self.name
                target_part = processor.func(arg=arg, lang=lang, base_grammar=base_grammar)
                result = result.replace(processor.match_string, target_part)
            self.value = result
            self.line = f'{self.name_with_priority}:{result}'

    def __str__(self):
        return f'{self.name}:{self.value}'

    def __repr__(self):
        return self.__str__()


class RuleProcessor:
    def __init__(self, name, arg):
        self.name = name
        self.arg = arg.strip()
        self.match_string = f'<{name}{arg}>'
        self.func = PREPROCESS_RULES[name]


def split_rule_name_and_value(s):
    # splits the name and the definition of a rule
    parts = s.split(':')
    if len(parts) <= 1:
        return s, s
    # we join because the rule definition could contain :
    return parts[0], ':'.join(parts[1:])


def apply_preprocessing_rules(grammar, base_grammar, lang):
    for rule in grammar.values():
        rule.apply_processors(base_grammar, lang)


def get_rule_from_grammar(rule_name, grammar):
    if rule_name not in grammar:
        raise Exception(f'There is a reference to rule {rule_name} but it is not in the base grammar.')
    return grammar[rule_name]


#
# Grammar rule preprocessing functions
#
def needs_colon(**kwargs):
    """ Returns the definition of the rule in the base grammar modified so that it is followed by a `:` """
    rule_name = kwargs['arg']
    base_grammar = kwargs['base_grammar']

    rule = get_rule_from_grammar(rule_name, base_grammar)
    value = rule.value
    pos = value.find('_EOL (_SPACE command)')
    return f'{value[0:pos]} _COLON {value[pos:]}'


def old_rule_to_error(**kwargs):
    """ Returns the 'old' version of the rule, i.e. the definition of the rule in the base grammar """
    arg = kwargs['arg']
    base_grammar = kwargs['base_grammar']

    rule = get_rule_from_grammar(arg, base_grammar)
    return rule.value


def expand_keyword(**kwargs):
    """ Creates a list of all values of a keyword. The keyword `else` produces `else|ellers` for Danish"""
    keyword = kwargs['arg']
    lang = kwargs['lang']

    values = get_translated_keyword(keyword, lang)
    values = sorted(list(set(values)))
    return '|'.join(values)


def expand_keyword_first(**kwargs):
    """ Creates a list of the first letter of all values of a keyword.
    The keyword `else` produces `ei` for Ukrainian """
    keyword = kwargs['arg']
    lang = kwargs['lang']

    values = get_translated_keyword(keyword, lang)
    values = sorted(list(set([v[0] for v in values])))
    return ''.join(values)


def expand_keyword_not_followed_by_space(**kwargs):
    """ Creates a negative lookahead for all values of a keyword (except their first letter) followed by a space.
    The keyword `else` produces `e(?!lse |llers )` for Danish and `e(?!lse )|и(?!наче)` for Bulgarian"""
    keyword = kwargs['arg']
    lang = kwargs['lang']

    values = get_translated_keyword(keyword, lang)
    first_to_rest = dict()
    for v in values:
        first, rest = v[0], f'{v[1:]} '
        if first in first_to_rest:
            first_to_rest[first].append(rest)
        else:
            first_to_rest[first] = [rest]

    result = [f'{k}(?!{"|".join(v)})' for k, v in first_to_rest.items()]
    return '|'.join(result)


def get_translated_keyword(keyword, lang):
    print(keyword, lang)

    def get_keyword_value_from_lang(keyword_, lang_):
        keywords = hedy_translation.keywords_to_dict(lang_)
        if keyword_ in keywords:
            return [k for k in keywords[keyword_] if k]
        else:
            raise Exception(f"The keywords yaml file for language '{lang_}' has no definition for '{keyword_}'.")

    translated_keyword = get_keyword_value_from_lang(keyword, lang) if lang != 'en' else []
    return translated_keyword + get_keyword_value_from_lang(keyword, 'en')


PREPROCESS_RULES = {
    'needs_colon': needs_colon,
    'old_rule_to_error': old_rule_to_error,
    'expand_keyword': expand_keyword,
    'expand_keyword_first': expand_keyword_first,
    'expand_keyword_not_followed_by_space': expand_keyword_not_followed_by_space,
}


#
# Grammar merging operators: +=, -=, >>
#
ADD_GRAMMAR_MERGE_OP = '+='
REMOVE_GRAMMAR_MERGE_OP = '-='
LAST_GRAMMAR_MERGE_OP = '>>'
GRAMMAR_MERGE_OPERATORS = [ADD_GRAMMAR_MERGE_OP, REMOVE_GRAMMAR_MERGE_OP, LAST_GRAMMAR_MERGE_OP]


def merge_rules_operator(prev_definition, new_definition, name, complete_line):
    op_to_arg = get_operator_to_argument(new_definition)

    add_arg = op_to_arg.get(ADD_GRAMMAR_MERGE_OP, '')
    remove_arg = op_to_arg.get(REMOVE_GRAMMAR_MERGE_OP, '')
    last_arg = op_to_arg.get(LAST_GRAMMAR_MERGE_OP, '')
    remaining_commands = get_remaining_rules(prev_definition, remove_arg, last_arg)
    ordered_commands = split_rule(remaining_commands, add_arg, last_arg)

    new_rule = f"{name}: {' | '.join(ordered_commands)}" if bool(op_to_arg) else complete_line
    deletable = split_rule(remove_arg)
    return new_rule, deletable


def get_operator_to_argument(definition):
    """Creates a map of all used operators and their respective arguments e.g. {'+=': 'print | play', '>>': 'echo'}"""
    operator_to_index = [(op, definition.find(op)) for op in GRAMMAR_MERGE_OPERATORS if op in definition]
    result = {}
    for i, (op, index) in enumerate(operator_to_index):
        start_index = index + len(op)
        if i + 1 < len(operator_to_index):
            _, next_index = operator_to_index[i + 1]
            result[op] = definition[start_index:next_index].strip()
        else:
            result[op] = definition[start_index:].strip()
    return result


def get_remaining_rules(orig_def, *sub_def):
    original_commands = split_rule(orig_def)
    commands_after_minus = split_rule(*sub_def)
    misses = [c for c in commands_after_minus if c not in original_commands]
    if misses:
        raise Exception(f"Command(s) {'|'.join(misses)} do not exist in the previous definition")
    remaining_commands = [cmd for cmd in original_commands if cmd not in commands_after_minus]
    remaining_commands = ' | '.join(remaining_commands)  # turn the result list into a string
    return remaining_commands


def split_rule(*rules):
    return [c.strip() for rule in rules for c in rule.split('|') if c.strip() != '']


def strip_priority_suffix(rule):
    if re.match(r"\w+\.-?\d+", rule):
        return rule.split('.')[0]
    return rule



import collections
import copy
import os
from ruamel import yaml


def extract_Lark_grammar_from_yaml():
    """Creates a lark file in ../grammars/ for  all yaml files located in ./data/keywords/.
    If a keyword is not yet translated, it will use the English translation of the keyword

    Args:
        only_new_lang (bool, optional): Specifies if only a lark file should be created for a new keyword language
        or for all languages. Defaults to True.
    """
    dirname = os.path.dirname(__file__)
    input_path = os.path.join(dirname, 'data', 'keywords')
    current_grammar_path = os.path.join(dirname, 'data', 'grammars')

    yaml_languages = [f.replace('.yaml', '') for f in os.listdir(input_path) if
                      os.path.isfile(os.path.join(input_path, f)) and f.endswith('.yaml')]

    template_lark = os.path.join(current_grammar_path, 'keywords-template.lark')
    with open(template_lark, 'r', encoding='utf-8') as f:
        template = f.read()

    for yaml_lang in yaml_languages:
        yaml_filesname_with_path = os.path.join(input_path, yaml_lang + '.yaml')
        default_yaml_with_path = os.path.join(input_path, 'en' + '.yaml')

        with open(default_yaml_with_path, 'r', encoding='utf-8') as stream:
            en_command_combinations = yaml.YAML(typ='safe', pure=True).load(stream)

        with open(yaml_filesname_with_path, 'r', encoding='utf-8') as stream:
            command_combinations = yaml.YAML(typ='safe', pure=True).load(stream)

        # Create an empty dictionary -> fill with english keywords and then overwrite all translated keywords
        translations = collections.defaultdict(lambda: 'Unknown Exception')
        translations.update(en_command_combinations)
        translations.update(command_combinations)

        translation_copy = copy.deepcopy(translations)
        for k, v in translation_copy.items():
            if yaml_lang == "ar":
                mixed_tatweel_in = ''.join([' "ـ"* ' + '"' + lang + '"' for lang in v]) + ' "ـ"* '
                translations[k] = mixed_tatweel_in
            else:
                # other languages need their translations surrounded by "'s
                translations[k] = '"' + v + '"'

            # we use | if we have multiple options, such as repete and repète
            if "|" in v:
                valid_translation = ""
                options = v.split("|")
                valid_translation = ' | '.join(['"' + option + '"' for option in options])
                translations[k] = valid_translation

        translated_template = template.format(**translations)

        lark_filesname_with_path = os.path.join(current_grammar_path, 'keywords-' + yaml_lang + '.lark')

        with open(lark_filesname_with_path, 'w', encoding='utf-8') as f:
            f.write(translated_template)


extract_Lark_grammar_from_yaml()



"""Functions that can be configured by the user of this library, such as the gettext function for translation."""

def initialize_gettext(gettext_fn):
    """Initialize the gettext function used for translation in this module."""
    global gettext
    gettext = gettext_fn


def gettext(s):
    """Default gettext function that just returns the input string."""
    return s


from os import path
import re
import textwrap

from lark import Tree

from . import exceptions
from .content import grammars_dir


class SourceRange:
    """
    A class used to represent source code ranges

    The source code range is made out of:
    from_line (int), from_column (int), to_line (int), to_column (int)

    An example:
    print Hello!
    ask What is your name?

    For the above snippet we could have the following mappings:
    print Hello! - from_line (1), from_column (1), to_line (1), to_column (13)
    ask What is your name? - from_line (2), from_column (1), to_line (2), to_column (23)
    Hello! - from_line (1), from_column (7), to_line (1), to_column (13)

    Tip: You can use a more advanced text editor like Notepad++ to get these values for a certain cursor position
    """

    def __init__(self, from_line, from_column, to_line, to_column):
        self.from_line = from_line
        self.from_column = from_column
        self.to_line = to_line
        self.to_column = to_column

    def __str__(self):
        return f'{self.from_line}/{self.from_column}-{self.to_line}/{self.to_column}'

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return (
            self.from_line, self.from_column,
            self.to_line, self.to_column
        ) == (
            other.from_line, other.from_column,
            other.to_line, other.to_column
        )


class SourceCode:
    """
    A class used to represent Hedy/Python source code

    The source code range is made out of:
    a source_range (SourceRange) and the code (str)
    """

    def __init__(self, source_range: SourceRange, code: str, error: Exception = None, command_name: str = None):
        self.source_range = source_range
        self.code = code
        self.error = error
        self.command_name = command_name

    def __hash__(self):
        return hash((
            self.source_range.from_line, self.source_range.from_column,
            self.source_range.to_line, self.source_range.to_column
        ))

    def __eq__(self, other):
        return (
            self.source_range.from_line, self.source_range.from_column,
            self.source_range.to_line, self.source_range.to_column
        ) == (
            other.source_range.from_line, other.source_range.from_column,
            other.source_range.to_line, other.source_range.to_column
        )

    def __ne__(self, other):
        return not (self == other)

    def __str__(self):
        if self.error is None:
            return f'{self.source_range} --- {self.code} --- {self.command_name}'
        else:
            return f'{self.source_range} -- ERROR[{self.error}] CODE[{self.code}]'

    def __repr__(self):
        return self.__str__()


def strip_priority_suffix(rule):
    if re.match(r"\w+\.\d+", rule):
        return rule.split('.')[0]
    return rule


class SourceMap:
    """
    A class used to represent the Hedy - Python source map.

    The map contains entries in the form of:
    SourceCode object (Hedy) - SourceCode object (Python)

    the string representation of the sourcemap is defined as:
    [Start line]-[Start Character]/[End line]-[End Character] :
        [Code]
    """

    map = dict()
    level = 0

    skip_faulty = False

    exceptions_not_to_skip = (
        exceptions.UnsupportedStringValue,
    )

    language = 'en'
    hedy_code = ''
    python_code = ''
    grammar_rules = []

    def __init__(self):
        self.get_grammar_rules()

    def set_level(self, level):
        self.level = level

    def set_language(self, language):
        self.language = language

    def set_skip_faulty(self, skip_faulty):
        # if the mapping encounters an error and skip_faulty is True we will 'skip' the exception
        self.skip_faulty = skip_faulty

    def set_hedy_input(self, hedy_code):
        self.hedy_code = hedy_code

    def set_python_output(self, python_code):
        self.python_code = python_code
        python_code_mapped = list()
        hedy_lines = self.hedy_code.split('\n')
        indent_size = find_program_indent_length(hedy_lines)

        def line_col(context, idx):
            return context.count('\n', 0, idx) + 1, idx - context.rfind('\n', 0, idx)

        for hedy_source_code, python_source_code in self.map.items():
            if hedy_source_code.error is not None or python_source_code.code == '':
                continue
            if self.level <= 7:
                start_index = -1
                number_of_indents = 0
                while start_index == -1 and number_of_indents < 10:
                    python_statement_code = textwrap.indent(python_source_code.code, '  '*number_of_indents)
                    start_index = python_code.find(python_statement_code)
                    code_char_length = len(python_source_code.code)

                    for i in range(python_code_mapped.count(python_source_code.code)):
                        start_index = python_code.find(python_statement_code, start_index+code_char_length)
                        start_index = max(0, start_index)  # not found (-1) means that start_index = 0
                    number_of_indents += 1
            else:
                number_of_indents = find_indent_length(
                    hedy_lines[hedy_source_code.source_range.from_line - 1]) // indent_size

                python_statement_code = textwrap.indent(python_source_code.code, '  '*number_of_indents)
                start_index = python_code.find(python_statement_code)
                code_char_length = len(python_source_code.code)

                for i in range(python_code_mapped.count(python_source_code.code)):
                    start_index = python_code.find(python_statement_code, start_index+code_char_length)
                    start_index = max(0, start_index)  # not found (-1) means that start_index = 0

            end_index = start_index + code_char_length
            start_line, start_column = line_col(python_code, start_index)
            end_line, end_column = line_col(python_code, end_index)

            python_source_code.source_range = SourceRange(
                start_line,
                start_column,
                end_line,
                end_column
            )

            python_code_mapped.append(python_source_code.code)

    def get_grammar_rules(self):
        with open(path.join(grammars_dir(), "level1.lark"), "r", encoding="utf-8") as file:
            grammar_text = file.read()

        for i in range(2, 19):
            with open(path.join(grammars_dir(), f'level{i}-Additions.lark'), "r", encoding="utf-8") as file:
                grammar_text += '\n' + file.read()

        self.grammar_rules = re.findall(r"([\w.]+):", grammar_text)
        self.grammar_rules = [rule for rule in self.grammar_rules if 'text' not in rule]  # exclude text from mapping
        self.grammar_rules = list(set(self.grammar_rules))  # remove duplicates
        self.grammar_rules = [strip_priority_suffix(r) for r in self.grammar_rules]

    def add_source(self, hedy_code: SourceCode, python_code: SourceCode):
        self.map[hedy_code] = python_code

    def clear(self):
        self.map.clear()
        self.level = 0
        self.language = 'en'
        self.hedy_code = ''
        self.python_code = ''

    def get_result(self):
        response_map = dict()
        index = 0

        for hedy_source_code, python_source_code in self.map.items():
            response_map[index] = {
                'hedy_range': {
                    'from_line': hedy_source_code.source_range.from_line,
                    'from_column': hedy_source_code.source_range.from_column,
                    'to_line': hedy_source_code.source_range.to_line,
                    'to_column': hedy_source_code.source_range.to_column,
                },
                'python_range': {
                    'from_line': python_source_code.source_range.from_line,
                    'from_column': python_source_code.source_range.from_column,
                    'to_line': python_source_code.source_range.to_line,
                    'to_column': python_source_code.source_range.to_column,
                },
                'error': hedy_source_code.error,
                'command': hedy_source_code.command_name
            }

            index += 1
        return response_map

    def get_compressed_mapping(self):
        response_map = dict()

        for hedy_source_code, python_source_code in self.map.items():
            response_map[str(hedy_source_code.source_range)] = str(python_source_code.source_range)

        return response_map

    def get_error_from_hedy_source_range(self, hedy_range: SourceRange) -> Exception:
        for hedy_source_code, python_source_code in self.map.items():
            if hedy_source_code.source_range == hedy_range:
                return hedy_source_code.error

    def print_source_map(self, d, indent=0):
        for key, value in d.items():
            print('\t' * indent + str(key) + ':')
            if isinstance(value, dict):
                self.print_source_map(value, indent + 1)
            else:
                code_lines = str(value).splitlines()
                code_lines = ['\t' * (indent + 1) + str(x) for x in code_lines]
                code = "\n".join(code_lines)
                print(code)
            print('')

    def __str__(self):
        self.print_source_map(self.map)
        return str()


def source_map_rule(source_map: SourceMap):
    """ A decorator function that should decorator the transformer method (grammar rule)
        the decorator adds the hedy code & python code to the map when the transformer method (grammar rule) is used
    """

    def decorator(function):
        def wrapper(*args, **kwargs):
            meta = args[1]

            hedy_code_input = source_map.hedy_code[meta.start_pos:meta.end_pos]
            hedy_code_input = hedy_code_input.replace('#ENDBLOCK', '')  # ENDBLOCK is not part of the Hedy code, remove
            error = None

            if not source_map.skip_faulty:
                generated_python = function(*args, **kwargs)
            else:
                try:
                    generated_python = function(*args, **kwargs)

                    # When parsing with skip_faulty enabled it could happen that because sanitization is not done
                    # a tree is returned instead of a string containing valid Python code by a transformer method.
                    # If this happens we have to raise an exception, we cannot map a Lark tree

                    if (
                        # if a Lark tree is returned
                        isinstance(generated_python, Tree) or
                        # if a Lark tree is returned as a string, we check with regex
                        bool(re.match(r".*Tree\(.*Token\(.*\).*\).*", generated_python))
                    ):
                        raise Exception('Can not map a Lark tree, only strings')

                except Exception as e:
                    # If an exception is found, we set the Python code to pass (null operator)
                    # we also map the error
                    generated_python = 'pass'
                    error = e

            hedy_code = SourceCode(
                SourceRange(
                    meta.container_line, meta.container_column,
                    meta.container_end_line, meta.container_end_column
                ),
                hedy_code_input,
                error=error,
                command_name=function.__name__
            )

            python_code = SourceCode(
                # We don't know now, set_python_output will set the ranges later
                SourceRange(None, None, None, None),
                generated_python
            )

            source_map.add_source(hedy_code, python_code)
            return generated_python

        return wrapper

    return decorator


def source_map_transformer(source_map: SourceMap):
    """ A decorator function that should decorate a transformer class

        This is used for convenience, instead of adding source_map_rule to all methods,
        source_map_transformer needs only to be added to the transformer class.
        This decorator add source_map_rule to all appropriate methods.
    """

    def decorate(cls):
        for rule in cls.__dict__:
            if rule in source_map.grammar_rules:
                setattr(cls, rule, source_map_rule(source_map)(getattr(cls, rule)))
        return cls

    return decorate


def find_indent_length(line):
    number_of_spaces = 0
    for x in line:
        if x == ' ':
            number_of_spaces += 1
        else:
            break
    return number_of_spaces


def find_program_indent_length(program_lines):
    indent_size = 4
    found_indent_size = False
    for line in program_lines:
        leading_spaces = find_indent_length(line)
        # continue if line is just spaces
        if leading_spaces == len(line):
            continue

        if not found_indent_size and leading_spaces > 0:
            indent_size = leading_spaces
            found_indent_size = True

    return indent_size



import os
from os import path
import contextlib

prefixes_dir = path.join(path.dirname(__file__), 'prefixes')

# Define code that will be used if some turtle command is present
with open(f'{prefixes_dir}/turtle.py', encoding='utf-8') as f:
    TURTLE_PREFIX_CODE = f.read()

# Preamble that will be used for non-Turtle programs
# numerals list generated from: https://replit.com/@mevrHermans/multilangnumerals
with open(f'{prefixes_dir}/normal.py', encoding='utf-8') as f:
    NORMAL_PREFIX_CODE = f.read()

# Define code that will be used if a pressed command is used
with open(f'{prefixes_dir}/pressed.py', encoding='utf-8') as f:
    PRESSSED_PREFIX_CODE = f.read()

# Define code that will be used if music code is used
with open(f'{prefixes_dir}/music.py', encoding='utf-8') as f:
    MUSIC_PREFIX_CODE = f.read()


def is_production():
    """Whether we are serving production traffic."""
    return os.getenv('IS_PRODUCTION', '') != ''


@contextlib.contextmanager
def atomic_write_file(filename, mode='wb'):
    """Write to a filename atomically.

    First write to a unique tempfile, then rename the tempfile into
    place. Use replace instead of rename to make it atomic on windows as well.
    Use as a context manager:

        with atomic_write_file('file.txt') as f:
            f.write('hello')
    """

    tmp_file = f'{filename}.{os.getpid()}'
    with open(tmp_file, mode) as f:
        yield f

    os.replace(tmp_file, filename)



from __future__ import annotations

import re
from collections import defaultdict, namedtuple

from lark import Token, Transformer, v_args
from lark.exceptions import VisitError

import hedy
from . import content as hedy_content

# Holds the token that needs to be translated, its line number, start and
# end indexes and its value (e.g. ", ").
Rule = namedtuple("Rule", "keyword line start end value")


def keywords_to_dict(lang="nl"):
    """Return a dictionary from keyword symbols to their translations in `lang`

    Examples:

    >>> keywords_to_dict('en')['not_in'] == ['not in']
    >>> keywords_to_dict('fr')['repeat'] == ['répète', 'repete']
    >>> keywords_to_dict('nl')['to'] == ['tot']
    >>> keywords_to_dict('nl')['to_list'] == ['toe ann']
    """
    return {
        # Sort the keywords by length descending. This is important for the substitution logic later
        k: list(sorted(local_keys, key=len, reverse=True))
        for k, local_keys in hedy_content.ALL_KEYWORDS[lang].items()
    }


def lang_switch_table(level: int, lang1: str, lang2: str | None = None):
    """Return a dictionary from keywords in lang1 organized in tuples to their translations in lang2, e.g.
        ('ask', ) -> ('vraag',)
        ('at', 'random') -> ('kies', 'willekeurig')
    """
    from_lang, to_lang = (lang1, lang2) if lang2 else ('en', lang1)

    phrases = [
        ('at', 'random'),
        ('add', 'to_list'),
        ('range', 'to'),
        ('remove', 'from'),
        ('repeat', 'times'),
    ]
    keysym_to_phrase = {k: phrase for phrase in phrases for k in phrase}

    def xlated_phrase(keysym, lang):
        phrase = keysym_to_phrase.get(keysym, (keysym,))
        return tuple(hedy_content.KEYWORDS[lang].get(k, k) for k in phrase)

    return {
        xlated_phrase(keyword_symbol, from_lang): xlated_phrase(keyword_symbol, to_lang)
        for keyword_symbol in hedy.commands_per_level[level]
    }


def all_keywords_to_dict():
    """Return a dictionary where each value is a list of the translations of that keyword (key). Used for testing"""
    keyword_dict = {}
    for lang in hedy_content.ALL_KEYWORD_LANGUAGES:
        keyword_dict[lang] = {k: v[0] for k, v in keywords_to_dict(lang).items()}

    all_translations = {k: [v.get(k, k) for v in keyword_dict.values()] for k in keyword_dict["en"]}
    return all_translations


def translate_keyword_from_en(keyword, lang="en"):
    # translated the keyword to a local lang
    local_keywords = keywords_to_dict(lang)
    if keyword in local_keywords.keys():
        local_keyword = local_keywords[keyword][0]
    else:
        local_keyword = keyword
    return local_keyword


def translate_keyword_to_en(keyword, lang):
    # translated the keyword to from a local lang
    original_keywords = keywords_to_dict(lang)
    for k, v in original_keywords.items():
        if keyword in v:
            return k
    return keyword


def get_target_keyword(keyword_dict, keyword):
    if keyword in keyword_dict.keys():
        return keyword_dict[keyword][0]
    else:
        return keyword


MATCH_EDGE_WHITESPACE = re.compile(r'^(\s*).+?(\s*)$')


def make_keyword_string_with_whitespace(matched_substring: str, new_keyword: str):
    """Make a new keyword string based on the matched substring.

    We retain all whitespace characters at the edge of the source string.
    """
    return MATCH_EDGE_WHITESPACE.sub(f'\\1{new_keyword}\\2', matched_substring)


def translate_keywords(input_string, from_lang="en", to_lang="nl", level=1):
    """ "Return code with keywords translated to language of choice in level of choice"""

    if input_string == "":
        return ""

    # remove leading spaces.
    # FH, dec 23. This creates a bit of a different version of translation but that seems ok to me
    # putting it back in seems overkill
    input_string = input_string.lstrip()

    try:
        processed_input = hedy.process_input_string(input_string, level, from_lang, preprocess_ifs_enabled=False)

        hedy.source_map.clear()
        hedy.source_map.set_skip_faulty(False)

        parser = hedy.get_parser(level, from_lang, True, hedy.source_map.skip_faulty)
        keyword_dict_from = keywords_to_dict(from_lang)
        keyword_dict_to = keywords_to_dict(to_lang)

        program_root = parser.parse(processed_input + "\n").children[0]

        translator = Translator(processed_input)
        translator.transform(program_root)

        # Build a list of textual substitutions, one per line
        # { line -> [(start, end, replacement)] }
        substitutions = defaultdict(list)

        lines = processed_input.splitlines()
        for rule in translator.rules:
            if rule.keyword in keyword_dict_from and rule.keyword in keyword_dict_to:
                # Sometimes the rule matches just a keyword, sometimes it matches a keyword
                # with spaces.
                line0 = rule.line - 1
                source_substring = lines[line0][rule.start:rule.end + 1]
                replaced_substring = make_keyword_string_with_whitespace(
                    source_substring, get_target_keyword(keyword_dict_to, rule.keyword))

                substitutions[line0].append((rule.start, rule.end + 1, replaced_substring))

        # Do the actual replacements, taking care to do them back-to-front
        for line0, subs in sorted(substitutions.items(), reverse=True):
            for start, end, replacement in sorted(subs, reverse=True):
                lines[line0] = lines[line0][:start] + replacement + lines[line0][end:]

        result = "\n".join(lines)
        result = result.replace("#ENDBLOCK", "")

        # we have to reverse escaping or translating and retranslating will add an unlimited number of slashes
        if level >= 4:
            result = result.replace("\\\\", "\\")

        return result
    except VisitError as E:
        if isinstance(E, VisitError):
            # Exceptions raised inside visitors are wrapped inside VisitError. Unwrap it if it is a
            # HedyException to show the intended error message.
            if isinstance(E.orig_exc, hedy.exceptions.HedyException):
                raise E.orig_exc
            else:
                raise E
    except Exception as E:
        raise E


def replace_line(lines, index, line):
    before = "\n".join(lines[0:index])
    after = "\n".join(lines[index + 1:])
    if len(before) > 0:
        before = before + "\n"
    if len(after) > 0:
        after = "\n" + after
    return "".join([before, line, after])


def replace_token_in_line(line, rule, original, target):
    """Replaces a token in a line from the user input with its translated equivalent"""
    before = "" if rule.start == 0 else line[0: rule.start]
    after = "" if rule.end == len(line) - 1 else line[rule.end + 1:]
    # Note that we need to replace the target value in the original value because some
    # grammar rules have ambiguous length and value, e.g. _COMMA: _SPACES*
    # (latin_comma | arabic_comma) _SPACES*
    return before + rule.value.replace(original, target) + after


def find_command_keywords(
    input_string, lang, level, keywords, start_line, end_line, start_column, end_column
):
    parser = hedy.get_parser(level, lang, True, hedy.source_map.skip_faulty)
    program_root = parser.parse(input_string).children[0]

    translator = Translator(input_string)
    translator.transform(program_root)

    return {
        k: find_keyword_in_rules(
            translator.rules, k, start_line, end_line, start_column, end_column
        )
        for k in keywords
    }


def find_keyword_in_rules(rules, keyword, start_line, end_line, start_column, end_column):
    for rule in rules:
        if rule.keyword == keyword and rule.line == start_line and rule.start >= start_column:
            if rule.line < end_line or (rule.line == end_line and rule.end <= end_column):
                return rule.value
    return None


@v_args(tree=True)
class Translator(Transformer):
    """The translator finds tokens that must be translated and stores information about their exact position
    in the user input string and original value. The information is later used to replace the token in
    the original user input with the translated token value. Please note that it is a transformer
    instead of a visitor because we need tokens to be visited too."""

    def __init__(self, input_string):
        super().__init__()
        self.input_string = input_string
        self.rules = []

    def define(self, tree):
        self.add_rule_for_grammar_token("_DEFINE", "define", tree)

    def defs(self, tree):
        self.add_rule_for_grammar_token("_DEF", "def", tree)

    def call(self, tree):
        self.add_rule_for_grammar_token("_CALL", "call", tree)

    def withs(self, tree):
        self.add_rule_for_grammar_token("_WITH", "with", tree)

    def returns(self, tree):
        self.add_rule_for_grammar_token("_RETURN", "return", tree)

    def print(self, tree):
        self.add_rule_for_grammar_token("_PRINT", "print", tree)

    def print_empty_brackets(self, tree):
        self.print(tree)

    def ask(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        self.add_rule_for_grammar_token("_ASK", "ask", tree)

    def echo(self, tree):
        self.add_rule_for_grammar_token("_ECHO", "echo", tree)

    def color(self, tree):
        self.add_rule_for_grammar_token("_COLOR", "color", tree)

    def forward(self, tree):
        self.add_rule_for_grammar_token("_FORWARD", "forward", tree)

    def turn(self, tree):
        self.add_rule_for_grammar_token("_TURN", "turn", tree)

    def left(self, tree):
        self.add_rule_for_grammar_rule("left", tree)

    def right(self, tree):
        self.add_rule_for_grammar_rule("right", tree)

    def black(self, tree):
        self.add_rule_for_grammar_rule("black", tree)

    def blue(self, tree):
        self.add_rule_for_grammar_rule("blue", tree)

    def brown(self, tree):
        self.add_rule_for_grammar_rule("brown", tree)

    def gray(self, tree):
        self.add_rule_for_grammar_rule("gray", tree)

    def green(self, tree):
        self.add_rule_for_grammar_rule("green", tree)

    def orange(self, tree):
        self.add_rule_for_grammar_rule("orange", tree)

    def pink(self, tree):
        self.add_rule_for_grammar_rule("pink", tree)

    def yellow(self, tree):
        self.add_rule_for_grammar_rule("yellow", tree)

    def purple(self, tree):
        self.add_rule_for_grammar_rule("purple", tree)

    def white(self, tree):
        self.add_rule_for_grammar_rule("white", tree)

    def red(self, tree):
        self.add_rule_for_grammar_rule("red", tree)

    def clear(self, tree):
        self.add_rule_for_grammar_rule("clear", tree)

    def TRUE(self, token):
        name = 'True' if token and token[0].isupper() else 'true'
        rule = Rule(name, token.line, token.column - 1, token.end_column - 2, token)
        self.rules.append(rule)

    def FALSE(self, token):
        name = 'False' if token and token[0].isupper() else 'false'
        rule = Rule(name, token.line, token.column - 1, token.end_column - 2, token)
        self.rules.append(rule)

    def assign_list(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        commas = self.get_keyword_tokens("_COMMA", tree)
        for comma in commas:
            rule = Rule("comma", comma.line, comma.column - 1, comma.end_column - 2, comma.value)
            self.rules.append(rule)

    def assign(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)

    def sleep(self, tree):
        self.add_rule_for_grammar_token("_SLEEP", "sleep", tree)

    def add(self, tree):
        self.add_rule_for_grammar_token("_ADD_LIST", "add", tree)
        self.add_rule_for_grammar_token("_TO_LIST", "to_list", tree)

    def remove(self, tree):
        self.add_rule_for_grammar_token("_REMOVE", "remove", tree)
        self.add_rule_for_grammar_token("_FROM", "from", tree)

    def random(self, tree):
        self.add_rule_for_grammar_rule("random", tree)

    def error_ask_dep_2(self, tree):
        self.add_rule_for_grammar_token("_ASK", "ask", tree)

    def error_echo_dep_2(self, tree):
        self.add_rule_for_grammar_token("_ECHO", "echo", tree)

    def if_clause(self, tree):
        self.add_rule_for_grammar_token("_IF", "if", tree)

    def ifelse(self, tree):
        self.add_rule_for_grammar_token("_IF", "if", tree)
        self.add_rule_for_grammar_token("_ELSE", "else", tree)

    def elif_clause(self, tree):
        self.add_rule_for_grammar_token("_ELIF", "elif", tree)

    def else_clause(self, tree):
        self.add_rule_for_grammar_token("_ELSE", "else", tree)

    def condition_spaces(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)

    def equality_check_is(self, tree):
        self.equality_check(tree)

    def equality_check(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        self.add_rule_for_grammar_token("_EQUALS", "=", tree)
        self.add_rule_for_grammar_token("_DOUBLE_EQUALS", "==", tree)

    def in_list_check(self, tree):
        self.add_rule_for_grammar_token("_IN", "in", tree)

    def list_access(self, tree):
        self.add_rule_for_grammar_token("_AT", "at", tree)

    def list_access_var(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        self.add_rule_for_grammar_token("_AT", "at", tree)

    def repeat(self, tree):
        self.add_rule_for_grammar_token("_REPEAT", "repeat", tree)
        self.add_rule_for_grammar_token("_TIMES", "times", tree)

    def for_list(self, tree):
        self.add_rule_for_grammar_token("_FOR", "for", tree)
        self.add_rule_for_grammar_token("_IN", "in", tree)

    def for_loop(self, tree):
        self.add_rule_for_grammar_token("_FOR", "for", tree)
        self.add_rule_for_grammar_token("_IN", "in", tree)
        self.add_rule_for_grammar_token("_RANGE", "range", tree)
        self.add_rule_for_grammar_token("_TO", "to", tree)

    def while_loop(self, tree):
        self.add_rule_for_grammar_token("_WHILE", "while", tree)

    def and_condition(self, tree):
        self.add_rule_for_grammar_token("_AND", "and", tree)

    def or_condition(self, tree):
        self.add_rule_for_grammar_token("_OR", "or", tree)

    def input(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        self.add_rule_for_grammar_token("_INPUT", "input", tree)

    def input_empty_brackets(self, tree):
        self.add_rule_for_grammar_token("_IS", "is", tree)
        self.add_rule_for_grammar_token("_INPUT", "input", tree)

    def pressed(self, tree):
        self.add_rule_for_grammar_token("_PRESSED", "pressed", tree)

    def add_rule_for_grammar_rule(self, rule_name, tree):
        """Creates a translation rule for a rule defined in the lark grammar which
        could have multiple children tokens, e.g. left, random, red"""
        # somehow for some Arabic rules (left, right, random) the parser returns separate tokens instead of one!
        token_start = tree.children[0]
        token_end = tree.children[-1]
        value = ''.join(tree.children)
        rule = Rule(rule_name, token_start.line, token_start.column - 1, token_end.end_column - 2, value)
        self.rules.append(rule)

    def add_rule_for_grammar_token(self, token_name, token_keyword, tree):
        """Creates a translation rule for a token defined in the lark grammar, e.g. _DEFINE, _FOR, _TURN"""
        token = self.get_keyword_token(token_name, tree)
        if token:
            rule = Rule(
                token_keyword, token.line, token.column - 1, token.end_column - 2, token.value
            )
            self.rules.append(rule)

    def get_keyword_token(self, token_type, node):
        for c in node.children:
            if type(c) is Token and c.type == token_type:
                return c
        return None

    def get_keyword_tokens(self, token_type, node):
        return [c for c in node.children if type(c) is Token and c.type == token_type]



from string import Formatter


def safe_format(s: str, /, **kwargs):
    """A replacement for str.format() that does NOT explode if values are missing from kwargs."""
    return FORMATTER.format(s, **kwargs)


class SafeFormatter(Formatter):
    def get_value(self, field_name, args, kwargs):
        if field_name not in kwargs:
            return '{' + str(field_name) + '}'
        return super().get_value(field_name, args, kwargs)


FORMATTER = SafeFormatter()



import inspect
"""
    Any exception added in this file must be also added to error-messages.txt
    So we can translate the error message. The exception must also be assigned
    an Exception Type in the exception_types dictionary in statistics.py
"""


class HedyException(Exception):
    def __init__(self, error_code, **arguments):
        """Create a new HedyException.

        You should not create a HedyException directly. Instead, use any of
        the subclasses of HedyException below.

        The keyword arguments passed into this constructor become available
        in exception translation strings. In those arguments, the keywords
        'location' and 'line_number' are special: they will be used to indicate
        the error location in the client.
        """
        super().__init__(error_code)

        self.error_code = error_code
        self.arguments = arguments

    @property
    def error_location(self):
        """Return the location where the error was found.

        Returns either an array of [row, col] or just [row].

        If 'location' is part of the keyword arguments, return that.
        Otherwise, if 'line_number' is part of the keyword arguments, return that instead
        wrapped in a list so we are sure the return type is always a list.

        """
        if 'location' in self.arguments:
            return self.arguments['location']
        if 'line_number' in self.arguments:
            return [self.arguments['line_number']]
        return None

    def __str__(self):
        return f'{self.error_code} {repr(self.arguments)}'


class WarningException(HedyException):
    """Fixed That For You warning/exception.

    Not really a failure case: instead it represents a warning to
    the user that they made a mistake we recovered for them.

    'fixed_code' and 'fixed_result' will contain the repaired
    code, and the result of compiling that repaired code.
    """

    def __init__(self, error_code, fixed_code, fixed_result, **arguments):
        super().__init__(error_code, **arguments)
        self.fixed_code = fixed_code
        self.fixed_result = fixed_result


class InvalidSpaceException(WarningException):
    def __init__(self, level, line_number, fixed_code, fixed_result):
        super().__init__('Invalid Space',
                         level=level,
                         line_number=line_number,
                         fixed_code=fixed_code,
                         fixed_result=fixed_result)  # what is the difference??


class UnusedVariableException(WarningException):
    def __init__(self, level, line_number, variable_name, fixed_code, fixed_result):
        super().__init__('Unused Variable',
                         level=level,
                         line_number=line_number,
                         variable_name=variable_name,
                         fixed_code=fixed_code,
                         fixed_result=fixed_result)


class ParseException(HedyException):
    def __init__(self, level, location, found, fixed_code=None):
        super().__init__('Parse',
                         level=level,
                         location=location,
                         found=found,
                         # 'character_found' for backwards compatibility
                         character_found=found)

        # TODO (FH, 8 dec 21) many exceptions now support fixed code maybe we
        # should move it to hedyexception?
        self.fixed_code = fixed_code


class UnquotedEqualityCheckException(HedyException):
    def __init__(self, line_number):
        super().__init__('Unquoted Equality Check',
                         line_number=line_number)
        self.location = [line_number]


class AccessBeforeAssignException(HedyException):
    def __init__(self, name, access_line_number, definition_line_number):
        super().__init__('Access Before Assign',
                         name=name,
                         access_line_number=access_line_number,
                         line_number=access_line_number,
                         definition_line_number=definition_line_number)


class UndefinedVarException(HedyException):
    def __init__(self, name, line_number):
        super().__init__('Var Undefined',
                         name=name,
                         line_number=line_number)


class UndefinedFunctionException(HedyException):
    def __init__(self, name, line_number):
        super().__init__('Function Undefined',
                         name=name,
                         line_number=line_number)


class CyclicVariableDefinitionException(HedyException):
    def __init__(self, variable, line_number):
        super().__init__('Cyclic Var Definition',
                         variable=variable,
                         line_number=line_number)


class InvalidArgumentTypeException(HedyException):
    def __init__(self, command, invalid_type, allowed_types, invalid_argument, line_number):
        super().__init__('Invalid Argument Type',
                         command=command,
                         invalid_type=invalid_type,
                         allowed_types=allowed_types,
                         invalid_argument=invalid_argument,
                         line_number=line_number)


class InvalidTypeCombinationException(HedyException):
    def __init__(self, command, arg1, arg2, type1, type2, line_number):
        super().__init__('Invalid Type Combination',
                         command=command,
                         invalid_argument=arg1,
                         invalid_argument_2=arg2,
                         invalid_type=type1,
                         invalid_type_2=type2,
                         line_number=line_number)


class InvalidArgumentException(HedyException):
    def __init__(self, command, allowed_types, invalid_argument, line_number):
        super().__init__('Invalid Argument',
                         command=command,
                         allowed_types=allowed_types,
                         invalid_argument=invalid_argument,
                         line_number=line_number)


class WrongLevelException(HedyException):
    def __init__(self, working_level, offending_keyword, tip, line_number):
        super().__init__('Wrong Level',
                         working_level=working_level,
                         offending_keyword=offending_keyword,
                         tip=tip,
                         line_number=line_number)


class InputTooBigException(HedyException):
    def __init__(self, lines_of_code, max_lines):
        super().__init__('Too Big',
                         lines_of_code=lines_of_code,
                         max_lines=max_lines)


class InvalidCommandException(WarningException):
    def __init__(
            self,
            level,
            invalid_command,
            guessed_command,
            line_number,
            fixed_code,
            fixed_result):
        super().__init__('Invalid',
                         invalid_command=invalid_command,
                         level=level,
                         guessed_command=guessed_command,
                         line_number=line_number,
                         fixed_code=fixed_code,
                         fixed_result=fixed_result)
        self.location = [line_number]


class MissingCommandException(HedyException):
    def __init__(self, level, line_number):
        super().__init__('Missing Command',
                         level=level,
                         line_number=line_number)


class MissingInnerCommandException(HedyException):
    def __init__(self, command, level, line_number):
        super().__init__('Missing Inner Command',
                         command=command,
                         level=level,
                         line_number=line_number)


class MissingVariableException(HedyException):
    def __init__(self, command, level, line_number):
        super().__init__('Missing Variable',
                         command=command,
                         level=level,
                         line_number=line_number)


class InvalidAtCommandException(HedyException):
    def __init__(self, command, level, line_number):
        super().__init__('Invalid At Command',
                         command=command,
                         level=level,
                         line_number=line_number)


class IncompleteRepeatException(HedyException):
    def __init__(self, command, level, line_number):
        super().__init__('Incomplete Repeat',
                         command=command,
                         level=level,
                         line_number=line_number)


class LonelyTextException(HedyException):
    def __init__(self, level, line_number):
        super().__init__('Lonely Text',
                         level=level,
                         line_number=line_number)


class IncompleteCommandException(HedyException):
    def __init__(self, incomplete_command, level, line_number):
        super().__init__('Incomplete',
                         incomplete_command=incomplete_command,
                         level=level,
                         line_number=line_number)

        # Location is copied here so that 'hedy_error_to_response' will find it
        # Location can be either [row, col] or just [row]
        self.location = [line_number]


class UnquotedTextException(HedyException):
    def __init__(self, level, line_number, unquotedtext=None):
        super().__init__('Unquoted Text',
                         level=level,
                         unquotedtext=unquotedtext,
                         line_number=line_number)


class MissingAdditionalCommand(HedyException):
    def __init__(self, command, missing_command, line_number):
        super().__init__('Missing Additional Command',
                         command=command,
                         missing_command=missing_command,
                         line_number=line_number)


class MisspelledAtCommand(HedyException):
    def __init__(self, command, arg1, line_number):
        super().__init__('Misspelled At Command',
                         command=command,
                         invalid_argument=arg1,
                         line_number=line_number)


class NonDecimalVariable(HedyException):
    def __init__(self, line_number):
        super().__init__('Non Decimal Variable',
                         line_number=line_number)


class UnquotedAssignTextException(HedyException):
    def __init__(self, text, line_number):
        super().__init__('Unquoted Assignment', text=text, line_number=line_number)


class MissingBracketsException(HedyException):
    def __init__(self, level, line_number):
        super().__init__('Missing Square Brackets',
                         line_number=line_number)


class LonelyEchoException(HedyException):
    def __init__(self):
        super().__init__('Lonely Echo')


class CodePlaceholdersPresentException(HedyException):
    def __init__(self, line_number):
        super().__init__('Has Blanks', line_number=line_number)


class TooManyIndentsStartLevelException(HedyException):
    def __init__(self, line_number, leading_spaces, fixed_code=None):
        super().__init__('Too Many Indents', line_number=line_number, leading_spaces=leading_spaces)
        self.fixed_code = fixed_code


class TooFewIndentsStartLevelException(HedyException):
    def __init__(self, line_number, leading_spaces, fixed_code=None):
        super().__init__('Too Few Indents', line_number=line_number, leading_spaces=leading_spaces)
        self.fixed_code = fixed_code


class NoIndentationException(HedyException):
    def __init__(self, line_number, leading_spaces, indent_size, fixed_code=None):
        super().__init__('No Indentation',
                         line_number=line_number,
                         leading_spaces=leading_spaces,
                         indent_size=indent_size)
        self.fixed_code = fixed_code


class IndentationException(HedyException):
    def __init__(self, line_number, leading_spaces, indent_size, fixed_code=None):
        super().__init__('Unexpected Indentation',
                         line_number=line_number,
                         leading_spaces=leading_spaces,
                         indent_size=indent_size)
        self.fixed_code = fixed_code


class UnsupportedFloatException(HedyException):
    def __init__(self, value):
        super().__init__('Unsupported Float', value=value)


class UnsupportedStringValue(HedyException):
    def __init__(self, invalid_value):
        super().__init__('Unsupported String Value', invalid_value=invalid_value)


class MissingElseForPressitException(HedyException):
    def __init__(self, command, level, line_number):
        super().__init__('Pressit Missing Else',
                         command=command,
                         level=level,
                         line_number=line_number)


class NestedFunctionException(HedyException):
    def __init__(self):
        super().__init__('Nested Function')


class WrongNumberofArguments(HedyException):
    def __init__(self, name, defined_number, used_number, line_number):
        super().__init__('Wrong Number of Arguments',
                         name=name,
                         defined_number=defined_number,
                         used_number=used_number,
                         line_number=line_number)


class InvalidErrorSkippedException(HedyException):
    def __init__(self):
        super().__init__('Invalid Error Skipped')


class RuntimeValueException(HedyException):
    def __init__(self, command, value, tip):
        super().__init__('Runtime Value Error', command=command, value=value, tip=tip)


class RuntimeValuesException(HedyException):
    def __init__(self, command, value, tip):
        super().__init__('Runtime Values Error', command=command, value=value, tip=tip)


class RuntimeIndexException(HedyException):
    def __init__(self, name):
        super().__init__('Runtime Index Error', name=name)


class ElseWithoutIfException(HedyException):
    def __init__(self, line_number):
        super().__init__('Else Without If Error', line_number=line_number)


class MissingColonException(HedyException):
    def __init__(self, command, line_number):
        super().__init__('Missing Colon Error', command=command, line_number=line_number)


HEDY_EXCEPTIONS = {name: cls for name, cls in globals().items() if inspect.isclass(cls)}



import hashlib
import logging
from os import path
import os
import pickle
import re
import tempfile
# Can't do this for now because of circular dependencies
# from ..website import querylog

from ruamel import yaml

from .lang_utils import atomic_write_file

yaml_loader = yaml.YAML(typ="safe", pure=True)
logger = logging.getLogger(__name__)

YAML_FILES_CACHE = {}


class YamlFile:
    """Data from a YAML file, accessible as if it is a dictionary.

    Use like so:

        file = YamlFile.for_file('path/to/file.yaml')

        if file.exists():
            print(file['key1'])
        else:
            print('oh no')

    Since loading the YAML files tends to be slow (takes 1s to load all YAML
    files for a single language in the course of rendering the main code editor page),
    we do some caching work:

    - Caches the loaded data in memory, on the Flask 'current request globals' object.
      This means that accessing the same file twice in the same request will only load
      it once. If we don't do this, loading YAMLs takes 2s instead of 1 because of the
      duplicate loads.
      - To keep the application memory footprint low, we don't cache the data
        permanently, but drop it after the request is done.
    - After we have successfully loaded a YAML file, we write a pickled version
      of that YAML file to disk, so that we can load the pickled version faster in
      the future future  (loading pickled data is ~400x faster than parsing a YAML
      file).
      - Pickled files are purely a server-side optimization. The files should not be checked
        in, and in fact by default this class will only generate them when running on Heroku.
      - Nevertheless, we do some time stamp checking to make sure to read the pickle
        file if its timestamp is older than the YAML file's timestamp.
    """

    @staticmethod
    def for_file(filename):
        """Factory function: return a singleton YamlFile instance for the given file."""
        filename = path.abspath(filename)
        if filename not in YAML_FILES_CACHE:
            YAML_FILES_CACHE[filename] = YamlFile(filename)
        return YAML_FILES_CACHE[filename]

    def __init__(self, filename):
        """Create a new YamlFile for the given filename."""
        self.filename = filename
        self.pickle_filename = path.join(tempfile.gettempdir(), 'hedy_pickles',
                                         f"{pathname_slug(self.filename)}.pickle")

    def exists(self):
        return os.path.exists(self.filename)

    def to_dict(self):
        """Return the contents of the file as a plain dict, or an empty dict if the file doesn't exist.

        You should generally not need to use this: this object can be used in places
        where dicts are expected (except if the file doesn't exist, then accessing it will throw; we
        can consider changing that behavior to always return an empty dict).
        """
        if self.exists():
            return self.access()
        return {}

    def access(self):
        """Access the data from memory.

        Load it if we haven't loaded it yet or the data on disk changed.
        """
        # Obtain or create a per-request cache dictionary (if we have a request), or an unattached
        # cache object that will disappear after this function returns if we don't have a request.
        yaml_cache = cache_object()
        cached = yaml_cache.get(self.filename)
        if cached is not None:
            return cached

        data = self.load()

        if not isinstance(data, dict):
            raise RuntimeError(f"Contents of {self.filename} needs to be a dict, got: {data}")

        yaml_cache[self.filename] = data
        return data

    def load(self):
        """Load the data from disk.

        Load from a pickle file if available, or load the original YAML
        and write a pickle file otherwise.
        """
        yaml_ts = self._file_timestamp(self.filename)
        pickle_ts = self._file_timestamp(self.pickle_filename)

        if pickle_ts and yaml_ts and pickle_ts > yaml_ts:
            # Pickle file is newer than the YAML, just read that
            return self.load_pickle()

        # Otherwise load uncached and save (atomically, since multiple processes might
        # be trying to write the pickle file in parallel)
        data = self.load_uncached()
        try:
            os.makedirs(path.dirname(self.pickle_filename), exist_ok=True)
            with atomic_write_file(self.pickle_filename) as f:
                pickle.dump(data, f)
        except IOError as e:
            logger.warning('Error writing pickled YAML: %s', e)

        return data

    # @querylog.timed_as('load_yaml_pickled')
    def load_pickle(self):
        with open(self.pickle_filename, "rb") as f:
            return pickle.load(f)

    # @querylog.timed_as('load_yaml_uncached')
    def load_uncached(self):
        """Load the source YAML file."""
        file = self._load_yaml(self.filename)
        base_filename = path.join(path.dirname(self.filename), 'en.yaml')
        if self.filename != base_filename:
            base_file = self._load_yaml(base_filename)
            merged = self.merge_yaml(file, base_file)
            return merged
        return file

    def _load_yaml(self, filename):
        try:
            with open(filename, 'r', encoding="utf-8") as f:
                return yaml_loader.load(f)
        except IOError:
            return {}

    @staticmethod
    def merge_yaml(source, fallback):
        """Merge the language file with the fallback file.

        The source yaml is merged with the fallback yaml, so that the values missing in the source are taken from
        the fallback. Note that if a mismatch occurs (the property is a list in the source but a dict in the fallback,
        or the source file has an array with more elements than its counterpart in the target file), the fallback
        content is preferred and the source content is discarded."""
        if source and isinstance(source, dict):
            if not isinstance(fallback, dict):
                return fallback
            for key, value in source.items():
                if key in fallback:
                    returned = YamlFile.merge_yaml(value, fallback.get(key))
                    fallback[key] = returned
        elif source and isinstance(source, list):
            if not isinstance(fallback, list):
                return fallback
            for i, value in enumerate(source):
                if i < len(fallback):
                    returned = YamlFile.merge_yaml(value, fallback[i])
                    fallback[i] = returned
        else:
            return source if source else fallback
        return fallback

    def _file_timestamp(self, filename):
        try:
            return os.stat(filename).st_mtime
        except FileNotFoundError:
            return None

    # Make this object look like a readonly 'dict'
    def __getitem__(self, key):
        return self.access()[key]

    def get(self, key, default=None):
        return self.access().get(key, default)

    def has_key(self, k):
        return k in self.access()

    def keys(self):
        return self.access().keys()

    def values(self):
        return self.access().values()

    def items(self):
        return self.access().items()

    def __contains__(self, item):
        return item in self.access()

    def __iter__(self):
        return iter(self.access())

    def __len__(self):
        return len(self.access())


def pathname_slug(x):
    """Turn a path name into an identifier we can use as a file name.

    Take into account that it may contain characters we want to remove,
    that the full path may be too long to use as a file name, and that
    it needs to remain unique.
    """
    x = re.sub(r'[^a-zA-Z0-9_-]', '', x)
    return x[-20:] + md5digest(x)


def md5digest(x):
    return hashlib.md5(x.encode("utf-8")).hexdigest()


def cache_object():
    """Return the cache object in which we will keep caches for YAML files.

    By default, will return an empty object, which means nothing will be cached.
    """
    return {}


def initialize_yaml_cache(cache_fn):
    """Replace the 'cache_object' function with the given function, which should return a cache object.

    For example, replace it with the following:

    ```py
    initializEe_yaml_cache(lambda: g.setdefault('yaml_cache', {}) if has_request_context() else {})
    ```
    """
    global cache_object
    cache_object = cache_fn


def insert(input_string, line, column, new_string):
    """"insert new_string at (line, column)"""
    rows = input_string.splitlines()
    rows[line] = rows[line][:column] + new_string + rows[line][column:]

    return '\n'.join(rows)


def delete(input_string, line, column, length):
    """"delete length chars starting at (line, column)"""
    rows = input_string.splitlines()
    rows[line] = rows[line][:column] + rows[line][column + length:]

    return '\n'.join(rows)


def replace(input_string, line, column, length, new_string):
    """"replace at (line, column) length chars with new_string"""
    result = delete(input_string, line, column, length)
    result = insert(result, line, column, new_string)

    return result


def remove_leading_spaces(input_string):
    # the only repair we can do now is remove leading spaces, more can be added!
    return '\n'.join([x.lstrip() for x in input_string.split('\n')])


def remove_unexpected_char(input_string, line, column):
    return delete(input_string, line - 1, column - 1, 1)


def fix_indent(input_string, line, leading_spaces, indent_size):
    if leading_spaces < indent_size:
        # not enough spaces, add spaces
        return insert(input_string, line - 1, 0, ' ' * (indent_size - leading_spaces))
    else:
        # too many spaces, remove spaces
        return delete(input_string, line - 1, 0, leading_spaces - indent_size)
